@@ -363,28 +363,25 @@ void ha_task_dump(struct buffer *buf, const struct task *task, const char *pfx)
 static int cli_io_handler_show_threads(struct appctx *appctx)
 {
 	struct stconn *sc = appctx_sc(appctx);
-	int thr;
+	int *thr = appctx->svcctx;
 
 	/* FIXME: Don't watch the other side !*/
 	if (unlikely(sc_opposite(sc)->flags & SC_FL_SHUT_DONE))
 		return 1;
 
-	if (appctx->st0)
-		thr = appctx->st1;
-	else
-		thr = 0;
+	if (!thr)
+		thr = applet_reserve_svcctx(appctx, sizeof(*thr));
 
 	do {
 		chunk_reset(&trash);
-		ha_thread_dump(&trash, thr);
+		ha_thread_dump(&trash, *thr);
 
 		if (applet_putchk(appctx, &trash) == -1) {
 			/* failed, try again */
-			appctx->st1 = thr;
 			return 0;
 		}
-		thr++;
-	} while (thr < global.nbthread);
+		(*thr)++;
+	} while (*thr < global.nbthread);
 
 	return 1;
 }
@@ -1470,6 +1467,8 @@ static int debug_iohandler_fd(struct appctx *appctx)
 
 		salen = sizeof(sa);
 		if (getsockname(fd, (struct sockaddr *)&sa, &salen) != -1) {
+			int i;
+
 			if (sa.ss_family == AF_INET)
 				port = ntohs(((const struct sockaddr_in *)&sa)->sin_port);
 			else if (sa.ss_family == AF_INET6)
@@ -1477,6 +1476,12 @@ static int debug_iohandler_fd(struct appctx *appctx)
 			else
 				port = 0;
 			addrstr = sa2str(&sa, port, 0);
+			/* cleanup the output */
+			for  (i = 0; i < strlen(addrstr); i++) {
+				if (iscntrl((unsigned char)addrstr[i]) || !isprint((unsigned char)addrstr[i]))
+					addrstr[i] = '.';
+			}
+
 			chunk_appendf(&trash, " laddr=%s", addrstr);
 			free(addrstr);
 		}
@@ -1490,6 +1495,11 @@ static int debug_iohandler_fd(struct appctx *appctx)
 			else
 				port = 0;
 			addrstr = sa2str(&sa, port, 0);
+			/* cleanup the output */
+			for  (i = 0; i < strlen(addrstr); i++) {
+				if ((iscntrl((unsigned char)addrstr[i])) || !isprint((unsigned char)addrstr[i]))
+					addrstr[i] = '.';
+			}
 			chunk_appendf(&trash, " raddr=%s", addrstr);
 			free(addrstr);
 		}
@@ -1726,7 +1736,6 @@ static void debug_release_memstats(struct appctx *appctx)
 void debug_handler(int sig, siginfo_t *si, void *arg)
 {
 	struct buffer *buf = HA_ATOMIC_LOAD(&th_ctx->thread_dump_buffer);
-	int harmless = is_thread_harmless();
 
 	/* first, let's check it's really for us and that we didn't just get
 	 * a spurious DEBUGSIG.
@@ -1738,13 +1747,6 @@ void debug_handler(int sig, siginfo_t *si, void *arg)
 	 * we come from a sig handler.
 	 */
 	ha_thread_dump_one(tid, 1);
-
-	/* mark the current thread as stuck to detect it upon next invocation
-	 * if it didn't move.
-	 */
-	if (!harmless &&
-	    !(_HA_ATOMIC_LOAD(&th_ctx->flags) & TH_FL_SLEEPING))
-		_HA_ATOMIC_OR(&th_ctx->flags, TH_FL_STUCK);
 }
 
 static int init_debug_per_thread()
