@@ -13,6 +13,7 @@
 #include <haproxy/qmux_http.h>
 #include <haproxy/qmux_trace.h>
 #include <haproxy/quic_conn.h>
+#include <haproxy/quic_enc.h>
 #include <haproxy/quic_frame.h>
 #include <haproxy/quic_sock.h>
 #include <haproxy/quic_stream.h>
@@ -100,14 +101,6 @@ static struct qcs *qcs_new(struct qcc *qcc, uint64_t id, enum qcs_type type)
 	qcs->st = QC_SS_IDLE;
 	qcs->ctx = NULL;
 
-	qcs->sd = sedesc_new();
-	if (!qcs->sd)
-		goto err;
-	qcs->sd->se   = qcs;
-	qcs->sd->conn = qcc->conn;
-	se_fl_set(qcs->sd, SE_FL_T_MUX | SE_FL_ORPHAN | SE_FL_NOT_FIRST);
-	se_expect_no_data(qcs->sd);
-
 	/* App callback attach may register the stream for http-request wait.
 	 * These fields must be initialed before.
 	 */
@@ -155,6 +148,14 @@ static struct qcs *qcs_new(struct qcc *qcc, uint64_t id, enum qcs_type type)
 
 	qcs->err = 0;
 
+	qcs->sd = sedesc_new();
+	if (!qcs->sd)
+		goto err;
+	qcs->sd->se   = qcs;
+	qcs->sd->conn = qcc->conn;
+	se_fl_set(qcs->sd, SE_FL_T_MUX | SE_FL_ORPHAN | SE_FL_NOT_FIRST);
+	se_expect_no_data(qcs->sd);
+
 	/* Allocate transport layer stream descriptor. Only needed for TX. */
 	if (!quic_stream_is_uni(id) || !quic_stream_is_remote(qcc, id)) {
 		struct quic_conn *qc = qcc->conn->handle.qc;
@@ -339,7 +340,7 @@ static void qcc_refresh_timeout(struct qcc *qcc)
 					/* We are past the soft close window end, wake the timeout
 					 * task up immediately.
 					 */
-					qcc->task->expire = now_ms;
+					qcc->task->expire = tick_add(now_ms, 0);
 					task_wakeup(qcc->task, TASK_WOKEN_TIMER);
 				}
 			}
@@ -627,6 +628,9 @@ static struct qcs *qcc_init_stream_remote(struct qcc *qcc, uint64_t id)
 	/* Only stream ID not already opened can be used. */
 	BUG_ON(id < *largest);
 
+	/* MAX_STREAMS emission must not allowed too big stream ID. */
+	BUG_ON(*largest > QUIC_VARINT_8_BYTE_MAX);
+
 	while (id >= *largest) {
 		const char *str = *largest < id ? "initializing intermediary remote stream" :
 		                                  "initializing remote stream";
@@ -683,9 +687,13 @@ struct stconn *qcs_attach_sc(struct qcs *qcs, struct buffer *buf, char fin)
 	BUG_ON_HOT(!LIST_INLIST(&qcs->el_opening));
 	LIST_DEL_INIT(&qcs->el_opening);
 
+	/* rcv_buf may be skipped if request is wholly received on attach.
+	 * Ensure that similar flags are set for FIN both on rcv_buf and here.
+	 */
 	if (fin) {
 		TRACE_STATE("report end-of-input", QMUX_EV_STRM_RECV, qcc->conn, qcs);
 		se_fl_set(qcs->sd, SE_FL_EOI);
+		se_expect_data(qcs->sd);
 	}
 
 	/* A QCS can be already locally closed before stream layer
@@ -1441,6 +1449,8 @@ int qcc_recv_stop_sending(struct qcc *qcc, uint64_t id, uint64_t err)
 	return 1;
 }
 
+#define QUIC_MAX_STREAMS_MAX_ID (1ULL<<60)
+
 /* Signal the closing of remote stream with id <id>. Flow-control for new
  * streams may be allocated for the peer if needed.
  */
@@ -1451,8 +1461,28 @@ static int qcc_release_remote_stream(struct qcc *qcc, uint64_t id)
 	TRACE_ENTER(QMUX_EV_QCS_END, qcc->conn);
 
 	if (quic_stream_is_bidi(id)) {
+		/* RFC 9000 4.6. Controlling Concurrency
+		 *
+		 * If a max_streams transport parameter or a MAX_STREAMS frame is
+		 * received with a value greater than 260, this would allow a maximum
+		 * stream ID that cannot be expressed as a variable-length integer; see
+		 * Section 16. If either is received, the connection MUST be closed
+		 * immediately with a connection error of type TRANSPORT_PARAMETER_ERROR
+		 * if the offending value was received in a transport parameter or of
+		 * type FRAME_ENCODING_ERROR if it was received in a frame; see Section
+		 * 10.2.
+		 */
+		if (qcc->lfctl.ms_bidi == QUIC_MAX_STREAMS_MAX_ID) {
+			TRACE_DATA("maximum streams value reached", QMUX_EV_QCC_SEND, qcc->conn);
+			goto out;
+		}
+
 		++qcc->lfctl.cl_bidi_r;
-		if (qcc->lfctl.cl_bidi_r > qcc->lfctl.ms_bidi_init / 2) {
+		/* MAX_STREAMS needed if closed streams value more than twice
+		 * the initial window or reaching the stream ID limit.
+		 */
+		if (qcc->lfctl.cl_bidi_r > qcc->lfctl.ms_bidi_init / 2 ||
+		    qcc->lfctl.cl_bidi_r + qcc->lfctl.ms_bidi == QUIC_MAX_STREAMS_MAX_ID) {
 			TRACE_DATA("increase max stream limit with MAX_STREAMS_BIDI", QMUX_EV_QCC_SEND, qcc->conn);
 			frm = qc_frm_alloc(QUIC_FT_MAX_STREAMS_BIDI);
 			if (!frm) {
@@ -1477,8 +1507,8 @@ static int qcc_release_remote_stream(struct qcc *qcc, uint64_t id)
 		 */
 	}
 
+ out:
 	TRACE_LEAVE(QMUX_EV_QCS_END, qcc->conn);
-
 	return 0;
 
  err:
@@ -1755,8 +1785,12 @@ void qcc_streams_sent_done(struct qcs *qcs, uint64_t data, uint64_t offset)
 		if (qcs->flags & (QC_SF_FIN_STREAM|QC_SF_DETACH)) {
 			/* Close stream locally. */
 			qcs_close_local(qcs);
-			/* Reset flag to not emit multiple FIN STREAM frames. */
-			qcs->flags &= ~QC_SF_FIN_STREAM;
+
+			if (qcs->flags & QC_SF_FIN_STREAM) {
+				qcs->stream->flags |= QC_SD_FL_WAIT_FOR_FIN;
+				/* Reset flag to not emit multiple FIN STREAM frames. */
+				qcs->flags &= ~QC_SF_FIN_STREAM;
+			}
 		}
 	}
 
@@ -2883,8 +2917,12 @@ static void qmux_strm_shutw(struct stconn *sc, enum co_shw_mode mode)
 	if (!qcs_is_close_local(qcs) &&
 	    !(qcs->flags & (QC_SF_FIN_STREAM|QC_SF_TO_RESET))) {
 
-		if (qcs->flags & QC_SF_UNKNOWN_PL_LENGTH) {
-			/* Close stream with a FIN STREAM frame. */
+		/* Close stream with FIN if length unknown and some data are
+		 * ready to be/already transmitted.
+		 * TODO select closure method on app proto layer
+		 */
+		if (qcs->flags & QC_SF_UNKNOWN_PL_LENGTH &&
+		    (qcs->tx.offset || b_data(&qcs->tx.buf))) {
 			if (!(qcc->flags & (QC_CF_ERR_CONN|QC_CF_ERRL))) {
 				TRACE_STATE("set FIN STREAM",
 				            QMUX_EV_STRM_SHUT, qcc->conn, qcs);
