@@ -22,9 +22,10 @@
 #include <haproxy/log.h>
 #include <haproxy/proxy.h>
 #include <haproxy/sample.h>
+#include <haproxy/sc_strm.h>
+#include <haproxy/stconn.h>
 #include <haproxy/stick_table.h>
 #include <haproxy/stream-t.h>
-#include <haproxy/stream_interface.h>
 #include <haproxy/tcp_rules.h>
 #include <haproxy/ticks.h>
 #include <haproxy/tools.h>
@@ -115,12 +116,12 @@ int tcp_inspect_request(struct stream *s, struct channel *req, int an_bit)
 	 * - if one rule returns KO, then return KO
 	 */
 
-	if ((req->flags & (CF_EOI|CF_SHUTR|CF_READ_ERROR)) || channel_full(req, global.tune.maxrewrite) ||
-	    si_rx_blocked_room(chn_prod(req)) ||
+	if ((s->scf->flags & (SC_FL_EOS|SC_FL_ABRT_DONE)) || channel_full(req, global.tune.maxrewrite) ||
+	    sc_waiting_room(s->scf) ||
 	    !s->be->tcp_req.inspect_delay || tick_is_expired(s->rules_exp, now_ms)) {
 		partial = SMP_OPT_FINAL;
 		/* Action may yield while the inspect_delay is not expired and there is no read error */
-		if ((req->flags & CF_READ_ERROR) || !s->be->tcp_req.inspect_delay || tick_is_expired(s->rules_exp, now_ms))
+		if ((s->scf->flags & SC_FL_ERROR) || !s->be->tcp_req.inspect_delay || tick_is_expired(s->rules_exp, now_ms))
 			act_opts |= ACT_OPT_FINAL;
 	}
 	else
@@ -163,6 +164,8 @@ resume_execution:
 						break;
 					case ACT_RET_STOP:
 					case ACT_RET_DONE:
+						s->last_rule_file = rule->conf.file;
+						s->last_rule_line = rule->conf.line;
 						goto end;
 					case ACT_RET_YIELD:
 						s->current_rule = rule;
@@ -170,16 +173,26 @@ resume_execution:
 							send_log(s->be, LOG_WARNING,
 								 "Internal error: yield not allowed if the inspect-delay expired "
 								 "for the tcp-request content actions.");
+							s->last_rule_file = rule->conf.file;
+							s->last_rule_line = rule->conf.line;
 							goto internal;
 						}
 						goto missing_data;
 					case ACT_RET_DENY:
+						s->last_rule_file = rule->conf.file;
+						s->last_rule_line = rule->conf.line;
 						goto deny;
 					case ACT_RET_ABRT:
+						s->last_rule_file = rule->conf.file;
+						s->last_rule_line = rule->conf.line;
 						goto abort;
 					case ACT_RET_ERR:
+						s->last_rule_file = rule->conf.file;
+						s->last_rule_line = rule->conf.line;
 						goto internal;
 					case ACT_RET_INV:
+						s->last_rule_file = rule->conf.file;
+						s->last_rule_line = rule->conf.line;
 						goto invalid;
 				}
 				continue; /* eval the next rule */
@@ -187,9 +200,13 @@ resume_execution:
 
 			/* If not action function defined, check for known actions */
 			if (rule->action == ACT_ACTION_ALLOW) {
+				s->last_rule_file = rule->conf.file;
+				s->last_rule_line = rule->conf.line;
 				goto end;
 			}
 			else if (rule->action == ACT_ACTION_DENY) {
+				s->last_rule_file = rule->conf.file;
+				s->last_rule_line = rule->conf.line;
 				goto deny;
 			}
 		}
@@ -239,9 +256,8 @@ resume_execution:
 		_HA_ATOMIC_INC(&sess->listener->counters->failed_req);
 
  reject:
-	si_must_kill_conn(chn_prod(req));
-	channel_abort(req);
-	channel_abort(&s->res);
+	sc_must_kill_conn(s->scf);
+	stream_abort(s);
 
  abort:
 	req->analysers &= AN_REQ_FLT_END;
@@ -284,12 +300,12 @@ int tcp_inspect_response(struct stream *s, struct channel *rep, int an_bit)
 	 * - if one rule returns OK, then return OK
 	 * - if one rule returns KO, then return KO
 	 */
-	if ((rep->flags & (CF_EOI|CF_SHUTR|CF_READ_ERROR)) || channel_full(rep, global.tune.maxrewrite) ||
-	    si_rx_blocked_room(chn_prod(rep)) ||
+	if ((s->scb->flags & (SC_FL_EOS|SC_FL_ABRT_DONE)) || channel_full(rep, global.tune.maxrewrite) ||
+	    sc_waiting_room(s->scb) ||
 	    !s->be->tcp_rep.inspect_delay || tick_is_expired(s->rules_exp, now_ms)) {
 		partial = SMP_OPT_FINAL;
 		/* Action may yield while the inspect_delay is not expired and there is no read error */
-		if ((rep->flags & CF_READ_ERROR) || !s->be->tcp_rep.inspect_delay || tick_is_expired(s->rules_exp, now_ms))
+		if ((s->scb->flags & SC_FL_ERROR) || !s->be->tcp_rep.inspect_delay || tick_is_expired(s->rules_exp, now_ms))
 			act_opts |= ACT_OPT_FINAL;
 	}
 	else
@@ -332,6 +348,8 @@ resume_execution:
 						break;
 					case ACT_RET_STOP:
 					case ACT_RET_DONE:
+						s->last_rule_file = rule->conf.file;
+						s->last_rule_line = rule->conf.line;
 						goto end;
 					case ACT_RET_YIELD:
 						s->current_rule = rule;
@@ -339,17 +357,27 @@ resume_execution:
 							send_log(s->be, LOG_WARNING,
 								 "Internal error: yield not allowed if the inspect-delay expired "
 								 "for the tcp-response content actions.");
+							s->last_rule_file = rule->conf.file;
+							s->last_rule_line = rule->conf.line;
 							goto internal;
 						}
 						channel_dont_close(rep);
 						goto missing_data;
 					case ACT_RET_DENY:
+						s->last_rule_file = rule->conf.file;
+						s->last_rule_line = rule->conf.line;
 						goto deny;
 					case ACT_RET_ABRT:
+						s->last_rule_file = rule->conf.file;
+						s->last_rule_line = rule->conf.line;
 						goto abort;
 					case ACT_RET_ERR:
+						s->last_rule_file = rule->conf.file;
+						s->last_rule_line = rule->conf.line;
 						goto internal;
 					case ACT_RET_INV:
+						s->last_rule_file = rule->conf.file;
+						s->last_rule_line = rule->conf.line;
 						goto invalid;
 				}
 				continue; /* eval the next rule */
@@ -357,16 +385,22 @@ resume_execution:
 
 			/* If not action function defined, check for known actions */
 			if (rule->action == ACT_ACTION_ALLOW) {
+				s->last_rule_file = rule->conf.file;
+				s->last_rule_line = rule->conf.line;
 				goto end;
 			}
 			else if (rule->action == ACT_ACTION_DENY) {
+				s->last_rule_file = rule->conf.file;
+				s->last_rule_line = rule->conf.line;
 				goto deny;
 			}
 			else if (rule->action == ACT_TCP_CLOSE) {
-				chn_prod(rep)->flags |= SI_FL_NOLINGER | SI_FL_NOHALF;
-				si_must_kill_conn(chn_prod(rep));
-				si_shutr(chn_prod(rep));
-				si_shutw(chn_prod(rep));
+				s->scb->flags |= SC_FL_NOLINGER | SC_FL_NOHALF;
+				sc_must_kill_conn(s->scb);
+				sc_abort(s->scb);
+				sc_shutdown(s->scb);
+				s->last_rule_file = rule->conf.file;
+				s->last_rule_line = rule->conf.line;
 				goto end;
 			}
 		}
@@ -421,9 +455,8 @@ resume_execution:
 		_HA_ATOMIC_INC(&__objt_server(s->target)->counters.failed_resp);
 
  reject:
-	si_must_kill_conn(chn_prod(rep));
-	channel_abort(rep);
-	channel_abort(&s->req);
+	sc_must_kill_conn(s->scb);
+	stream_abort(s);
 
   abort:
 	rep->analysers &= AN_RES_FLT_END;
@@ -442,7 +475,7 @@ resume_execution:
 /* This function performs the TCP layer4 analysis on the current request. It
  * returns 0 if a reject rule matches, otherwise 1 if either an accept rule
  * matches or if no more rule matches. It can only use rules which don't need
- * any data. This only works on connection-based client-facing stream interfaces.
+ * any data. This only works on connection-based client-facing stream connectors.
  */
 int tcp_exec_l4_rules(struct session *sess)
 {
@@ -539,7 +572,7 @@ int tcp_exec_l4_rules(struct session *sess)
 /* This function performs the TCP layer5 analysis on the current request. It
  * returns 0 if a reject rule matches, otherwise 1 if either an accept rule
  * matches or if no more rule matches. It can only use rules which don't need
- * any data. This only works on session-based client-facing stream interfaces.
+ * any data. This only works on session-based client-facing stream connectors.
  * An example of valid use case is to track a stick-counter on the source
  * address extracted from the proxy protocol.
  */
@@ -1027,7 +1060,7 @@ static int tcp_parse_request_rule(char **args, int arg, int section_type,
 			memprintf(err,
 			          "'%s %s' expects 'accept', 'reject', 'capture', 'expect-proxy', 'expect-netscaler-cip', 'track-sc0' ... 'track-sc%d', %s "
 			          "in %s '%s' (got '%s').%s%s%s\n",
-			          args[0], args[1], MAX_SESS_STKCTR-1,
+			          args[0], args[1], global.tune.nb_stk_ctr-1,
 			          trash.area, proxy_type_str(curpx),
 			          curpx->id, args[arg],
 			          best ? " Did you mean '" : "",
@@ -1163,7 +1196,7 @@ static int tcp_parse_tcp_rep(char **args, int section_type, struct proxy *curpx,
 
 	return warn;
  error:
-	free(rule);
+	free_act_rule(rule);
 	return -1;
 }
 
@@ -1377,7 +1410,7 @@ static int tcp_parse_tcp_req(char **args, int section_type, struct proxy *curpx,
 
 	return warn;
  error:
-	free(rule);
+	free_act_rule(rule);
 	return -1;
 }
 
